local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local map = require("map")
local exploration = require("campaign/exploration")

local path = require("scenario/pseudoRandomTour/path")
local levels = require("scenario/pseudoRandomTour/levels")
local vehicle = require("scenario/pseudoRandomTour/vehicle")
local waypoints = require("scenario/pseudoRandomTour/waypoints")
local tableComp = require("scenario/pseudoRandomTour/tableComp")
local uiSettings = require("scenario/pseudoRandomTour/uiSettings")
local achievements = require("scenario/pseudoRandomTour/achievements")
local levelSettings = require("scenario/pseudoRandomTour/levelSettings")
local configurations = require("scenario/pseudoRandomTour/configurations")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local lastQuery = 0
local showRoadIndicator = false
local showDestinationOnly = false

local markerColour

local isPRTScenario = false

----------------------------------------------------------------------------------------------------

local function generateRouteMarkerColour()
  --[[
  Generates a random route colour
  Returns:
      <table> - the RGB route colour
  ]]--
  local colours = {}
  local black = true

  for i = 1, 3 do
    -- Only generate easy to see colours, so the channel should be empty or full
    local colour = math.random(0, 1)
    colours[i] = colour

    if colour > 0 then
      black = false
    end
  end

  -- Don't have a black path colour
  if black then
    -- Set one channel to be full
    colours[math.random(1, 3)] = 1
  end

  return colours
end

local function freeze(freezeVeh)
  --[[
  Freezes all cars in the scenario
  ]]--
  scenario_scenarios.freezeAll(freezeVeh and 1 or 0)
end

local function defaultTargetIndicator()
  --[[
  Sets up the target indicator in its default, empty setting
  ]]--
  exploration.state = {
    closestRoad = {
      player = { first = nil, second = nil, best = nil, pos = nil },
      target = { first = nil, second = nil, best = nil, pos = nil }
    },
    focusExplorationPOI = false,
    focusExplorationScenario = false
  }

  showRoadIndicator = true
end

local function removeTargetIndicator()
  --[[
  Removes and resets the target indicator
  ]]--
  showRoadIndicator = false
  exploration.state = nil
  core_groundMarkers.resetAll()
end

local function setupRoadIndicator(show)
  --[[
  Sets up, or removes, the road indicator
  Parameters:
      show - whether to show the road indicator
  ]]--
  if show then
    defaultTargetIndicator()
  else
    removeTargetIndicator()
  end
end

local function reset()
  --[[
  Resets to the initial state
  ]]--
  removeTargetIndicator()
  path.setRoute(nil)
  waypoints.resetWaypoints()
  guihooks.trigger("prt_reset")
end

local function apply(routeGenerated, routeLength, firstNode, secondNode, alwaysSunny, addTraffic)
  --[[
  Applies and generates the scenario based on the route generated
  Parameters:
      routeGenerated - the route generated
      routeLength    - the length of the route generated
      firstNode      - the first node's data for the path
      secondNode     - the second node's data for the path
      alwaysSunny    - whether the time of day is always sunny or not
      addTraffic     - whether to generate traffic or not
  ]]--
  local scenario = scenario_scenarios.getScenario()

  -- No branching paths so the configs are the same
  scenario.BranchLapConfig = scenario.lapConfig
  scenario_waypoints.initialise()

  levelSettings.setupRandom(alwaysSunny)

  path.setRoute(routeGenerated)
  path.setCurrentRouteLength(routeLength)

  if addTraffic then
    gameplay_traffic.setupTraffic(nil, 0, 0)
  else
    gameplay_traffic.deleteVehicles()
  end

  local firstNodePos = firstNode.pos
  -- Reset UI and show next two waypoints
  guihooks.trigger("prt_reset")
  waypoints.showWaypointOnUi(
      waypoints.getLastWaypointData(firstNode.name) and "last" or "next",
      firstNode.name,
      firstNode.scale,
      firstNodePos.x,
      firstNodePos.y
  )

  -- Possibly only one waypoint
  if secondNode then
    local secondNodePos = secondNode.pos
    waypoints.showWaypointOnUi(
        "nextNext",
        secondNode.name,
        secondNode.scale,
        secondNodePos.x,
        secondNodePos.y
    )
  end

  markerColour = generateRouteMarkerColour()
end

local function jsBoolToLua(jsBool)
  return jsBool == true or jsBool == 1
end

local function initialise(routeLength, waypointSpacing, alwaysSunny, addTraffic, useRoadIndicator, destinationOnly, keepCurrentVehicle)
  --[[
  Setups the scenario for the player
  Parameters:
      routeLength        - the target length of the route
      waypointSpacing    - the spacing of the waypoints (default, sparse, sparser)
      alwaysSunny        - whether the time of day is always sunny or not
      addTraffic         - whether to allow traffic or not
      useRoadIndicator   - whether to show the road indicator or not
      destinationOnly    - whether to only show the destination waypoint
      keepCurrentVehicle - whether to keep the current vehicle
  Returns:
      <string> - the status (ERR, WARN or OK)
      <string> - (only if status = ERR or WARN) - the description of what went wrong/unexpected
      <number> - the length of the route generated (m)
      <number> - the number of waypoints created
      <number> - the average driveability of the route
  ]]--
  -- Compatibility with JavaScript
  alwaysSunny = jsBoolToLua(alwaysSunny)
  addTraffic = jsBoolToLua(addTraffic)
  useRoadIndicator = jsBoolToLua(useRoadIndicator)
  destinationOnly = jsBoolToLua(destinationOnly)
  keepCurrentVehicle = jsBoolToLua(keepCurrentVehicle)

  setupRoadIndicator(useRoadIndicator)
  showDestinationOnly = destinationOnly

  uiSettings.setSettings(routeLength, alwaysSunny, addTraffic, useRoadIndicator, destinationOnly, keepCurrentVehicle, waypointSpacing)

  local scenario = scenario_scenarios.getScenario()

  -- Generate the route
  local gp = map.getGraphpath()
  local routeGenerated, length = path.generatePath(gp, routeLength)

  if not routeGenerated then
    return { "ERR", "No suitable path found for this length!" }
  end

  local avgDriveability = path.getRouteDriveability(gp, routeGenerated)
  path.setCurrentRouteDriveability(avgDriveability)

  -- Only keep the necessary nodes
  routeGenerated = path.trimRouteNodes(routeGenerated, gp, waypoints.getWaypointSpacingValue(waypointSpacing))
  -- Add in the scalings of each node for the waypoints
  routeGenerated = path.addNodeScalings(routeGenerated)

  local firstWp = routeGenerated[1].pos

  -- Recalculate the distance, since it is literately 'as the crow flies' distance
  if destinationOnly then
    length = math.floor((routeGenerated[#routeGenerated].pos - firstWp):length())
  end


  -- Waypoint nodes are the same, except remove the first to ensure the player does not trigger
  -- the first waypoint on spawn
  local waypointNodes = tableComp.mergeAppend(routeGenerated, {})
  table.remove(waypointNodes, 1);
  -- Generate the waypoints from the nodes
  waypoints.createWaypointsFromNodes(scenario, waypointNodes, destinationOnly)

  -- Place the player on the first node, pointing at the second
  vehicle.placePlayer(
      avgDriveability,
      vehicle.getPlayerAngleDirection(firstWp, routeGenerated[2].pos),
      firstWp,
      not keepCurrentVehicle
  )

  -- Set up the scenario for the route
  apply(
      routeGenerated,
      length,
      waypoints.getWaypointData("node_1"),
      waypoints.getWaypointData("node_2"),
      alwaysSunny,
      addTraffic
  )

  return { "OK", length, #waypoints.getWaypointsInScenario(), math.floor(avgDriveability * 100) / 100 }
end

local function start()
  --[[
  Start the scenario
  ]]--
  scenario_scenarios.changeState("running")

  -- Show the road indicator
  local player = scenetree.findObject("scenario_player0")
  exploration.queryRoadNodeToPosition(player, player:getPosition(), "player")
  exploration.queryRoadNodeToPosition(player, waypoints.getWaypointData("node_1").pos, "target")

  -- Initialise the waypoint UI
  guihooks.trigger('WayPointChange', { current = 0, count = #waypoints.getWaypointsInScenario() })
end

local function save(filename, routeLength)
  --[[
  Saves a route to file
  Parameters:
      filename    - the name of the file to save the route as
      routeLength - the length of the route
  ]]--
  local additionalInfo
  if showDestinationOnly then
    local route = path.getRoute()
    local startPos = route[1].pos
    -- Store the player pos and angle, since these won't be possible to extrapolate from one waypoint
    additionalInfo = {
      playerPos = {
        x = startPos.x,
        y = startPos.y,
        z = startPos.z
      },
      playerAngle = vehicle.getPlayerAngleDirection(startPos, route[2].pos)
    }
  end

  configurations.save(
      levels.getMapName(),
      filename,
      routeLength,
      path.getCurrentRouteDriveability(),
      waypoints.getWaypointsInScenario(),
      additionalInfo
  )

  return { "OK" }
end

local function load(filename, alwaysSunny, addTraffic, useRoadIndicator, keepCurrentVehicle)
  --[[
  Load a route from file
  Parameters:
      filename - the name of the file
  Returns:
      <table> - if an error occurred, returns status "ERR" and the error description
                else, returns status "OK", the route length and the number waypoints loaded
  ]]--
  local fileContents = configurations.load(levels.getMapName(), filename)

  -- Can't find the route, respond with error
  if not fileContents then
    return { "ERR", "This file does not exist" }
  end

  -- Compatibility with JavaScript
  alwaysSunny = jsBoolToLua(alwaysSunny)
  addTraffic = jsBoolToLua(addTraffic)
  useRoadIndicator = jsBoolToLua(useRoadIndicator)
  keepCurrentVehicle = jsBoolToLua(keepCurrentVehicle)

  setupRoadIndicator(useRoadIndicator)

  local additionalInfo = fileContents.additionalInfo
  local wps = fileContents.waypoints
  local firstWp, secondWp = wps[1], wps[2]

  waypoints.createWaypointsFromNodes(scenario_scenarios.getScenario(), wps)

  if additionalInfo then
    vehicle.placePlayer(
        fileContents.routeDriveability,
        additionalInfo.playerAngle,
        vec3(additionalInfo.playerPos),
        not keepCurrentVehicle
    )
  else
    vehicle.placePlayer(
        fileContents.routeDriveability,
        vehicle.getPlayerAngleDirection(vec3(firstWp.pos), vec3(secondWp.pos)),
        firstWp.pos,
        not keepCurrentVehicle
    )
  end

  apply(wps, fileContents.routeLength, firstWp, secondWp, alwaysSunny, addTraffic)

  return { "OK", fileContents.routeLength, #waypoints.getWaypointsInScenario() }
end

local function previousRouteExists()
  --[[
  Returns:
      <table<int,boolean>> - a table of whether a route has been generated to play and whether a
                             route has been generated in the past
  ]]--
  return { path.getRoute() ~= nil, #waypoints.getWaypointsInScenario() > 0 }
end

local function getCurrentRoute()
  --[[
  Returns:
      <table(length, num waypoints)> - the length (m) and the number of waypoints in the currently
                                       loaded route
  ]]--
  if previousRouteExists()[1] then
    return { waypoints.getLengthOfCurrentRoute(), #waypoints.getWaypointsInScenario() }
  end
end

local function isPlayingScenario()
  --[[
  Returns:
      <boolean> - whether the scenario is being played
  ]]--
  return scenario_scenarios.getScenario().state == "running"
end

local function findExistingRoutes()
  --[[
  Returns:
      <table<string>> - a list of routes saved for this map
  ]]--
  local mapName = levels.getMapName()
  local routes = configurations.findExistingFiles(mapName)

  for i, route in ipairs(routes) do
    routes[i] = route:gsub("/prt/"..mapName.."/", "")
  end

  return routes
end


----------------------------------------------------------------------------------------------------
-- BeamNG Functions
----------------------------------------------------------------------------------------------------

local function onScenarioFinished()
  if not isPRTScenario then
    return
  end

  achievements.addScenarioToStats(
      path.getCurrentRouteLength() / 1000,
      #waypoints.getWaypointsInScenario(),
      vehicle.getVehicleConfig()
  )
end

local function onRaceWaypointReached(data)
  if not isPRTScenario then
    return
  end


  -- Remove waypoint reached and next waypoint (for updating colour)
  guihooks.trigger("prt_reset")

  -- Get the next two waypoints
  local nextWpData = waypoints.getNextWaypointData(data.waypointName)
  local nextNextWpData = nextWpData and waypoints.getNextWaypointData(nextWpData.name)

  -- Set this waypoint as passed
  waypoints.setLastPassedWaypoint(tonumber(data.waypointName:match("%d+")))

  if nextWpData then
    local nextWpPos = nextWpData.pos

    if showRoadIndicator then
      -- Set the next waypoint as the next target
      exploration.queryRoadNodeToPosition(
          scenetree.findObject("scenario_player0"),
          nextWpPos,
          "target"
      )
    end

    -- Add next waypoint to UI
    waypoints.showWaypointOnUi(
        waypoints.getLastWaypointData(nextWpData.name) and "last" or "next",
        nextWpData.name,
        nextWpData.scale,
        nextWpPos.x,
        nextWpPos.y
    )

    -- Show the next next waypoint as greyed out (like in actual scenarios)
    if nextNextWpData then
      local nextNextPos = nextNextWpData.pos
      waypoints.showWaypointOnUi(
          "nextNext",
          nextNextWpData.name,
          nextNextWpData.scale,
          nextNextPos.x,
          nextNextPos.y
      )
    end
  else
    -- Deactivate target indicator since this is the last waypoint
    removeTargetIndicator()
  end
end

local function onPreRender(dt)
  if not isPRTScenario then
    return
  end

  local state = exploration.state

  if not state or not state.closestRoad then
    return
  end


  -- See exploration.onPreRender() --

  lastQuery = lastQuery - dt
  if lastQuery <= 0 then
    local pVehicle = scenetree.findObject("scenario_player0")
    if not pVehicle then
      return
    end

    local vehiclePos = pVehicle:getPosition()
    exploration.queryRoadNodeToPosition(pVehicle, vehiclePos, 'player')
    lastQuery = 2.0
  end

  local playerRoadData = state.closestRoad['player']
  if playerRoadData and playerRoadData.first and playerRoadData.first ~= 'nil' then
    for key, data in pairs(state.closestRoad) do
      if key ~= 'player' and data.best then
        -- Set a random colour
        core_groundMarkers.setFocus(data.best, nil, nil, nil, nil, nil, markerColour)
      end
    end
  end
end

local function onScenarioRestarted()
  if not isPRTScenario then
    return
  end

  reset()
end

local function onScenarioUIReady()
  if not isPRTScenario then
    return
  end

  guihooks.trigger('MenuHide')
  --guihooks.trigger('ChangeState', 'menu')
end

local function onScenarioChange(sc)
  if sc then
    if not isPRTScenario then
      local scFile = sc.sourceFile
      if scFile and scFile:find("pseudoRandomTour") then
        isPRTScenario = true
      else
        -- Not a PRT scenario, end here
        return
      end
    end

    if sc.state == "pre-running" then
      -- Hide start screen
      guihooks.trigger("MenuHide")
      --guihooks.trigger("ChangeState", "menu")
    elseif sc.state == "running" then
      -- Force a quick start - don't wait for countdown
      sc.countDownTime = 0
      freeze(false)
    end
  end
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.initialise = initialise
M.start = start

-- Main UI interface
M.getMapNameFormatted = levels.getMapNameFormatted
M.save = save
M.load = load
M.previousRouteExists = previousRouteExists
M.getCurrentRoute = getCurrentRoute
M.getAchievementStats = achievements.getAchievementStats
M.findNewModdedMaps = levels.findNewModdedMaps
M.generateScenario = levels.copyScenarioFilesTo
M.loadLastUISettings = uiSettings.getSettings
M.isPlayingScenario = isPlayingScenario
M.findExistingRoutes = findExistingRoutes
M.freeze = freeze

-- BeamNG interface
M.onScenarioFinished = onScenarioFinished
M.onScenarioRestarted = onScenarioRestarted
M.onRaceWaypointReached = onRaceWaypointReached
M.onPreRender = onPreRender
M.onScenarioChange = onScenarioChange
M.onScenarioUIReady = onScenarioUIReady
return M