local M = {}

local function mergeAppend(a, b, ...)
  --[[
  Merges two or more integer indexed tables into a new table, preventing overwriting
  of the same index
  Parameters:
    a   - the first table
    b   - the second table
    ... - any additional tables
  Returns:
    c - the merged table
  ]]--
  local tables = { a, b, ... }

  local c = {}
  local index = 1

  for _, T in ipairs(tables) do
    for _, v in ipairs(T) do
      c[index] = v
      index = index + 1
    end
  end

  return c
end

local function getKeys(T)
  --[[
  Parameters:
      T - the dictionary
  Returns:
      <table> - the list of keys
  ]]--
  local keys = {}

  for k, _ in pairs(T) do
    table.insert(keys, k)
  end

  return keys
end

local function getRandomKey(T)
  --[[
  Parameters:
      T - the dictionary
  Returns:
      <?> - a random key in {T}
  ]]--
  local keys = getKeys(T)
  return keys[math.random(1, #keys)]
end

local function getRandomEntry(T)
  --[[
  Parameters:
      T - the dictionary
  Returns:
      <?> - a random entry in {T}
  ]]--
  local keys = getKeys(T)
  -- Select random key
  local key = keys[math.random(1, #keys)]
  return T[key]
end

local function getRandomWaypointEntry(T)
  --[[
  Parameters:
      T - the dictionary
  Returns:
      <?> - a random waypoint in {T}
  ]]--
  for k, _ in pairs(T) do
    if not k:find("DecalRoad") then
      return k
    end
  end
end

local function getTableFromEntry(T, k, v)
  --[[
  Returns the table containing a key and value pair of a nested table
  This function is recursively defined
  Parameters:
    T - the table
    k - the key of the value to find
    v - the value associated with the key
  Returns:
    <table> - the table found
  ]]--
  for key, value in pairs(T) do
    if type(value) == "table" then
      local t = getTableFromEntry(value, k, v)

      if t then
        return t
      end
    end

    if key == k and value == v then
      return T
    end
  end
end

local function lengthOf(T)
  --[[
  Parameters:
      T - the dictionary
  Returns:
      <number> - the length of the dictionary
  ]]--
  local c = 0

  for _, _ in pairs(T) do
    c = c + 1
  end

  return c
end

local function repeatTableFrom(T, originalT, n)
  --[[
  A helper function for the function duplicate, which adds the
  contents in originalT to T n times
  Parameters:
    T - the current table
    originalT - the original table
    n - the number of times to repeat the table
  Returns:
    <table> - the repeated table T
  ]]--
  if n == 1 then
    return T
  end

  return repeatTableFrom(mergeAppend(T, originalT), originalT, n - 1)
end

local function repeatTable(T, n)
  --[[
  Repeats the entries in a table T n times
  Uses a helper function to provide a simpler interface
  Parameters:
    T - the table
    n - the number of times to repeat
  Parameters:
    <table> - the repeated table T
  ]]--
  return repeatTableFrom(T, T, n)
end

local function split(T, n, m)
  --[[
  Splits a table between two indexes
  Parameters:
      T - the table
      n - the starting index (inclusive) - can be nil for default of 1
      m - the ending index (inclusive) - can be nil for default of the length of the table
  ]]--
  -- Defaults
  T = T or {}
  n = n or 1
  m = m or #T

  -- Assert valid parameters (also, n == 1 and m == #T is the same as T)
  if #T == 0 or n < m or (n == 1 and m == #T) then
    return T
  end

  local subT = {}

  for i, v in ipairs(T) do
    if i > n then
      break
    end

    if i >= m then
      table.insert(subT, v)
    end
  end

  return subT
end

local function contains(T, v)
  --[[
  Returns whether a table T contains a value v
  This function is recursively defined
  Parameters:
    T - the table
    v - the value to find
  Returns:
    <boolean> - whether T contains v
  ]]--
  for _, value in ipairs(T) do
    if type(value) == "table" then
      if contains(value, v) then
        return true
      end
    else
      if value == v then
        return true
      end
    end
  end

  return false
end

M.mergeAppend = mergeAppend
M.getKeys = getKeys
M.getRandomWaypointEntry = getRandomWaypointEntry
M.getRandomKey = getRandomKey
M.getRandomEntry = getRandomEntry
M.getTableFromEntry = getTableFromEntry
M.lengthOf = lengthOf
M.repeatTable = repeatTable
M.split = split
M.contains = contains
return M