local M = {}

----------------------------------------------------------------------------------------------------
-- Imports
----------------------------------------------------------------------------------------------------

local multispawn = require("core/multiSpawn")

local tableComp = require("scenario/pseudoRandomTour/tableComp")

----------------------------------------------------------------------------------------------------

local function randomisePlayerConfig(routeDriveability)
  --[[
  Randomises the player's car, car colour and configuration of car
  The off-road ability of the car will be taking into account for off-road paths
  Parameters:
      routeDriveability = the driveability of the route the player will take
  ]]--
  local possibleConfigs = multispawn.getInstalledVehicleData({ allMods = true, allConfigs = true })

  local chosenVehicle, chosenPartConfig, chosenPartInfo

  while true do
    -- Choose a random configuration
    local chosenVehConfigs = possibleConfigs[math.random(1, #possibleConfigs)]

    -- Get the random configuration model and part configuration
    local partConfig = chosenVehConfigs.configData[math.random(1, #chosenVehConfigs.configData)].config
    chosenVehicle = chosenVehConfigs.model

    chosenPartConfig = "vehicles/"..chosenVehicle.."/"..partConfig..".pc"
    chosenPartInfo = jsonReadFile("vehicles/"..chosenVehicle.."/info_"..partConfig..".json")

    -- If we don't know, assume it is so-so offroad
    local vehicleOffRoadScore = (chosenPartInfo and chosenPartInfo["Off-Road Score"] or 50) / 100

    --[[If the route driveability and vehicle offroad score are the same, this vehicle will
        be picked. Otherwise, give it a random chance, weighted based on how suitable it is]]--
    local probability = 1 - (routeDriveability - vehicleOffRoadScore)

    if math.random() <= probability then
      break
    end
  end

  local possiblePaints = jsonReadFile("vehicles/"..chosenVehicle.."/info.json").paints
  local selectedPaint = tableComp.getRandomEntry(possiblePaints)

  core_vehicles.replaceVehicle(chosenVehicle, {
    config = chosenPartConfig,
    paint = selectedPaint
  })
end

local function getVehicleHeight(vehObj)
  --[[
  Parameters:
      vehObj - the vehicle object
  Returns:
      <number> - the height of the vehicle
  ]]--
  local lowestNode = math.huge
  local sumOfNodes = vec3(0, 0, 0)

  local nodeCount = vehObj:getNodeCount()

  for i = 0, nodeCount do
    local pos = vehObj:getNodePosition(i)

    if pos.z > lowestNode then
      lowestNode = pos.z
    end

    if pos.z < lowestNode then
      lowestNode = pos.z
    end
  end

  local center = sumOfNodes / nodeCount

  return center.z - (lowestNode / 2)
end

local function getPlayerAngleDirection(startPos, nextPos)
  --[[
  Returns the player angle direction; from the first waypoint, facing the direction of the second
  waypoint
  Parameters:
      startPos - the position the player will be placed at
      nextPos  - the position the player will point to
  Returns:
      <number> - the player angle
  ]]--
  local direction = nextPos - startPos
  -- Position to second waypoint
  return math.deg(math.atan2(direction.x, direction.y)) + 180
end

local function placePlayer(routeDriveability, angle, pos, randomiseConfig)
  --[[
  Places the player at the first waypoint, facing the direction of the second waypoint
  Parameters:
      routeDriveability - the driveability of the route the player will take
      angle             - the angle of the player vehicle
      randomiseConfig   - whether to randomise the player configuration
  ]]--
  -- +0.4 for extra tolerance for bridges
  local vehicleHeight = getVehicleHeight(scenetree.findObject("scenario_player0")) + 0.4

  --[[Place the vehicle at the first waypoint, above by half the height of the vehicle to avoid
      clipping in the ground]]--
  TorqueScript.eval('scenario_player0.position = "'..pos.x..' '..pos.y..' '..(pos.z + vehicleHeight / 2)..'";')

  if randomiseConfig then
    randomisePlayerConfig(routeDriveability)
  end

  TorqueScript.eval('scenario_player0.rotation = "0 0 1 '..angle..'";')
end

local function getVehicleConfig()
  --[[
  Returns:
      <string> - the vehicle configuration and model, formatted as <vehicle>/<config>
  ]]--
  return scenetree.findObject("scenario_player0").partConfig:gsub(".pc", ""):gsub("vehicles/", "")
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.getPlayerAngleDirection = getPlayerAngleDirection
M.placePlayer = placePlayer
M.getVehicleConfig = getVehicleConfig
return M