local M = {}

local path        = require("scenario/pseudoRandomTour/path")
local race_marker = require("scenario/race_marker")

----------------------------------------------------------------------------------------------------
-- Variables
----------------------------------------------------------------------------------------------------

local wpsInScenario = {}
local waypointSpacing = 0
local lastPassedWaypoint = ""

----------------------------------------------------------------------------------------------------

local function showWaypointOnUi(type, name, scale, x, y)
    --[[
    Shows a waypoint on the map UI
    Parameters:
        type  - the waypoint type (next, nextNext or last)
        name  - the name of the waypoint
        scale - the waypoint's scale
        x     - the x coordinate of the waypoint
        y     - the y coordinate of the waypoint
    ]]--
    --guihooks.trigger("prt_show_waypoint", {type = type, wpName = name, scale = scale, x = x, y = y})
    --guihooks.trigger("NavigationStaticMarkers", { key = "PRT_waypoints", items })
end

local function emptyCurrentWaypoints()
    --[[
    Removes all waypoints in the scenario from the wpsInScenario variable
    ]]--
    wpsInScenario = {}
end

--- Deletes all objects of the form [namePrefix..i]. This will continue until a object of the next
--- index cannot be found
--- @param namePrefix (string) the prefix of the name of the objects to delete
--- @param i (number) the starting index to begin at
--- @param descend (boolean) whether to decrement the index instead of incrementing
local function deleteObjs(namePrefix, i, descend)
    local obj = scenetree.findObject(namePrefix..i)

    while obj do
        obj:delete()

        i = descend and i-1 or i+1
        obj = scenetree.findObject(namePrefix..i)
    end
end

--- Version 0.22.1 doesn't delete checkpoint markers automatically anymore, so delete manually
local function deleteMarkers()
    local markers = {"base_marker_", "cylinder_marker_left", "cylinder_marker_right"}
    -- Get current ID as this lingers and is very difficult to keep track of, since it is neither
    -- exposed in race_marker.lua, nor reset when the scenario is fully reloaded
    local nextID = race_marker.createRaceMarker().id

    for _, marker in ipairs(markers) do
        deleteObjs(marker, nextID - 1, true)
    end
end

local function resetWaypoints()
    --[[
    Delete all waypoints previously created
    Also empties the lap and branch configs in the scenario
    ]]--
    deleteObjs("node_", 1, false)
    deleteMarkers()

    -- Reset configs
    local scenario = scenario_scenarios.getScenario()
    scenario.lapConfig = {}
    scenario.BranchLapConfig = {}
end

local function createWaypoint(scenario, waypointName, position, scale, onlyWaypoint)
    --[[
    Creates a waypoint at a given position
    Parameters:
        scenario     - the scenario instance
        waypointName - the name of the waypoint
        position     - the position of the waypoint
        scale        - the scale of the waypoint
        onlyWaypoint - whether this is the ONLY waypoint that will be created
    ]]--
    -- Delete for re-creation
    if scenetree.findObject(waypointName) then
        TorqueScript.eval(waypointName..".delete();")
    end

    -- Make this waypoint very tall to be visible from any point on the map, if it is the only one
    local zScale = onlyWaypoint and 100 or scale

    TorqueScript.eval([[
        new BeamNGWaypoint(]]..waypointName..[[) {
          position = "]]..position.x..[[ ]]..position.y..[[ ]]..position.z..[[";
          scale = "]]..scale..[[ ]]..scale..[[ ]]..zScale..[[";
        };

        ScenarioObjectsGroup.add(]]..waypointName..[[);
    ]])

    -- Set the node data for the waypoint
    scenario.nodes[waypointName] = {
        pos = position,
        radius = scale
    }

    -- Add it to the lap-config for it to be registered
    table.insert(scenario.lapConfig, waypointName)

    table.insert(wpsInScenario, {
        name = waypointName,
        pos = {
            x = position.x,
            y = position.y,
            z = position.z
        },
        scale = scale
    })
end

local function createWaypointsFromNodes(scenario, nodes, destinationOnly)
    --[[
    Creates waypoints from a list of nodes (named node_[1->#nodes])
    Parameters:
        scenario        - the scenario instance
        nodes           - the list of node (should contain a position vector and a scale value)
        destinationOnly - whether to only show the last checkpoint
    ]]--
    resetWaypoints()
    emptyCurrentWaypoints()

    if destinationOnly then
        local node = nodes[#nodes]
        -- Give the waypoint a very tall height to be visible from any point in a map
        createWaypoint(scenario, "node_1", vec3(node.pos), node.scale, true)
    else
        for i, data in ipairs(nodes) do
            if i ~= 1 then
                createWaypoint(scenario, "node_"..(i-1), vec3(data.pos), data.scale, false)
            end
        end
    end
end

local function findWaypointsInLevel()
    --[[
    Returns:
        <table> - the list of waypoint names in the scenario
    ]]--
    return scenetree.findClassObjects("BeamNGWaypoint")
end

local function getNextWaypointData(currentName)
    --[[
    Parameters:
        currentName - the current waypoint name
    Returns:
        <table> - the next waypoint's data (nil if there is no next waypoint)
    ]]--
    for i, data in ipairs(wpsInScenario) do
        if data.name == currentName and i < #wpsInScenario then
            return wpsInScenario[i+1]
        end
    end
end

local function getLastWaypointData(currWaypointName)
    --[[
    Parameters:
        currWaypointName - the waypoint to check
    Returns:
        <boolean> - whether this waypoint is the last waypoint
    ]]--
    return wpsInScenario[#wpsInScenario].name == currWaypointName
end

local function getWaypointData(waypointName)
    --[[
    Parameters:
        waypointName - the name of the waypoint
    Returns:
        <table> - the waypoint data
    ]]--
    for _, data in ipairs(wpsInScenario) do
        if data.name == waypointName then
            return data
        end
    end

    print("No data exists for waypoint ", waypointName)
end

local function setLastPassedWaypoint(waypointNum)
    --[[
    Parameters:
        waypointNum - the waypoint number
    ]]--
    lastPassedWaypoint = waypointNum
end

local function getWaypointSpacingValue(wpSpacing)
    --[[
    Parameters:
        wpSpacing - the type of spacing to use (default, sparse, sparser)
    Returns:
        <number> - the waypoint spacing (in m)
    ]]--
    if wpSpacing == "sparse" then
        return 250
    elseif wpSpacing == "sparser" then
        return 500
    else
        return 100
    end
end

local function getLastPassedWaypoint()
    return lastPassedWaypoint
end

local function getLengthOfCurrentRoute()
    --[[
    Returns:
        <number> - the length of the currently loaded route (in m)
    ]]--
    return path.getLengthOfRoute(wpsInScenario)
end

local function getWaypointSpacing()
    return waypointSpacing
end

local function getWaypointsInScenario()
    return wpsInScenario
end

local function resetIDs()
    --nextWaypointId = 1
end

----------------------------------------------------------------------------------------------------
-- Interface
----------------------------------------------------------------------------------------------------

M.showWaypointOnUi = showWaypointOnUi
M.createWaypointsFromNodes = createWaypointsFromNodes
M.findWaypointsInLevel = findWaypointsInLevel
M.getNextWaypointData = getNextWaypointData
M.resetWaypoints = resetWaypoints
M.getLastWaypointData = getLastWaypointData
M.getWaypointData = getWaypointData
M.setLastPassedWaypoint = setLastPassedWaypoint
M.getWaypointSpacingValue = getWaypointSpacingValue
M.getLastPassedWaypoint = getLastPassedWaypoint
M.getLengthOfCurrentRoute = getLengthOfCurrentRoute
M.emptyCurrentWaypoints = emptyCurrentWaypoints
M.getWaypointSpacing = getWaypointSpacing
M.getWaypointsInScenario = getWaypointsInScenario
M.resetIDs = resetIDs
return M